#lang racket/base
;; Virtualized operating system, this time with presence.

(require racket/set)
(require racket/match)
(require (only-in racket/list flatten))
(require "unify.rkt")

(provide nested-vm
	 ground-vm

	 (struct-out topic)
	 topic-publisher
	 topic-subscriber
	 co-roles
	 co-topics
	 topic-union

	 (struct-out handlers)

	 (except-out (struct-out transition) transition)
	 (rename-out [make-transition transition])
	 extend-transition
	 prefix-transition

	 role
	 role/fresh
	 (except-out (struct-out add-role) add-role)
	 (rename-out [make-add-role add-role])
	 (except-out (struct-out delete-role) delete-role)
	 (rename-out [make-delete-role delete-role])
	 (except-out (struct-out send-message) send-message)
	 (rename-out [make-send-message send-message])
	 (except-out (struct-out spawn) spawn)
	 (rename-out [make-spawn spawn])
	 (except-out (struct-out kill) kill)
	 (rename-out [make-kill kill])
	 (except-out (struct-out yield) yield)
	 (rename-out [yield-macro yield])

	 (struct-out at-meta-level)

	 ;; Aliases of structures themselves, because of shadowing of
	 ;; constructors/type-names.
	 (rename-out [topic <topic>])
	 (rename-out [handlers <handlers>])
	 (rename-out [transition <transition>])
	 (rename-out [add-role <add-role>])
	 (rename-out [delete-role <delete-role>])
	 (rename-out [send-message <send-message>])
	 (rename-out [spawn <spawn>])
	 (rename-out [kill <kill>])
	 (rename-out [yield <yield>])
	 (rename-out [at-meta-level <at-meta-level>])

	 ;; Reexports from unify.rkt for convenience
	 wild
	 wild?
	 non-wild?

	 ;; Reexports from racket/match for convenience
	 (all-from-out racket/match)

	 ;; For debugging
	 current-ground-transition)

;; Endpoints are the units of deduplication.
;; Flows (in canonical form) are the units of presence.

;;---------------------------------------------------------------------------
;; Data definitions

;; A PID is an (arbitrary) VM-unique process identifier. Concretely,
;; it's an integer.

;; A PreEID is an arbitrary, process-chosen-and-supplied identifier
;; for an endpoint. It is to be equal?-comparable. It is to be unique
;; within the scope of a single process.

;; A EID is an (arbitrary) VM-unique endpoint identifier. Concretely,
;; it's an (eid PID PreEID). As a consequence of the scope of PreEIDs,
;; EIDs shouldn't be visible outside the scope of the owning process.
(struct eid (pid pre-eid) #:transparent)

;; One endpoint, one topic (which may be changed over time!), with the
;; caveat that as we are at present unable to represent true topic
;; unions, we actually store a *set* of topics against each
;; endpoint. The (current!) topic for the endpoint is to be taken as
;; the union of all the members in the set.

;; A Flow is a Topic that comes from the intersection of two dual
;; topics.

;; A sent message includes a "body" and a "role", and is equivalent to
;; a non-virtual topic with that role and with the given "body" as a
;; pattern. In a sense, topics quite literally are patterns over
;; entire messages.

;; A QuasiQueue<X> is a list of Xs in *reversed* order.

(struct vm (name ;; Any - for debugging complex VM trees
	    processes ;; Hash<PID, Process>
	    endpoints ;; Hash<EID, Endpoint>
	    next-process-id ;; PID
	    pending-actions ;; QuasiQueue<(cons PID Action)>
	    ) #:transparent)

;; An Endpoint is an (endpoint EID Set<Topic> Handlers), representing a
;; facet of a process responsible for playing a particular role (the
;; topic) in a conversation.
(struct endpoint (id topics handlers) #:transparent)

;; A Process is an Exists State . (process Any PID State Set<EID>
;; Set<EID>), representing a VM process and its collection of active
;; endpoints at this level and at the VM's container's level.
(struct process (name id state endpoints meta-endpoints) #:transparent)

;; A Topic is a (topic Role Pattern Boolean), describing an Endpoint's
;; role in a conversation.
(struct topic (role pattern virtual?) #:prefab)

;; BootK = either (PID -> Transition) or Transition
;; InterruptK = State -> Transition
;; TrapK<X> = X -> InterruptK

;; (handlers Maybe<TrapK<Topic>>
;;           Maybe<TrapK<Topic * Reason>>
;;           Maybe<TrapK<Topic * Message>>)
(struct handlers (presence absence message) #:transparent)

;; actions is a plain old ordered ConsTreeOf<Action>, not a
;; QuasiQueue.
(struct transition (state actions) #:transparent)

;; Preactions.
;; Ks are various TrapKs or #f, signifying lack of interest.
;;
;; (add-role PreEID (or Topic Set<Topic>) Handlers)
(struct add-role (pre-eid topics handlers) #:prefab)
;;
;; (delete-role PreEID Any)
(struct delete-role (pre-eid reason) #:prefab)
;;
;; (send-message Any Role)
(struct send-message (body role) #:prefab)
;;
;; (spawn BootK Maybe<TrapK<PID>> Any)
(struct spawn (main k debug-name) #:prefab)
;;
;; (kill Maybe<PID> Any)
(struct kill (pid reason) #:prefab)

;; An Action is either a Preaction or a (yield InterruptK) or an
;; (at-meta-level Preaction).
(struct yield (k) #:prefab)
(struct at-meta-level (preaction) #:prefab)

;;---------------------------------------------------------------------------
;; role & yield macros

(require (for-syntax syntax/parse))
(require (for-syntax racket/base))

(define-syntax role
  (lambda (stx)
    (syntax-parse stx
      [(_ pre-eid topics-expr
	  #:state state-pattern
	  (~or (~optional (~seq #:on-presence presence) #:name "#:on-presence handler")
	       (~optional (~seq #:on-absence absence) #:name "#:on-absence handler")
	       (~optional (~seq #:topic topic) #:defaults ([topic #'t0]) #:name "#:topic")
	       (~optional (~seq #:reason reason) #:defaults ([reason #'r0]) #:name "#:reason"))
	  ...
	  [message-pattern clause-body ...]
	  ...)
       (define-syntax-rule (build-handler args e-attr)
	 (if (not (attribute e-attr))
	     #'#f
	     #`(lambda args (match-lambda [state-pattern e-attr]))))
       (with-syntax ([presence-handler (build-handler (topic) presence)]
		     [absence-handler (build-handler (topic reason) absence)]
		     [message-handler #'(lambda (topic message-body)
					  (lambda (state)
					    (match state
					      [state-pattern
					       (match message-body
						 [message-pattern clause-body ...]
						 ...
						 [_ state])])))])
	 #'(add-role pre-eid
		     topics-expr
		     (handlers presence-handler absence-handler message-handler)))])))

(define-syntax-rule (role/fresh pre-eid-var rest ...)
  (let ((pre-eid-var (gensym 'role)))
    (role pre-eid-var rest ...)))

(define-syntax-rule (yield-macro #:state state-pattern body ...)
  (yield (match-lambda [state-pattern body ...])))

;;---------------------------------------------------------------------------
;; Smarter constructors for transitions and preactions.

(define (make-transition state . actions) (transition state actions))
(define make-add-role add-role) ;; no special treatment required at present
(define (make-delete-role pre-eid [reason #f]) (delete-role pre-eid reason))
(define (make-send-message body [role 'publisher]) (send-message body role))
(define (make-spawn main [k #f] #:debug-name [debug-name #f]) (spawn main k debug-name))
(define (make-kill [pid #f] #:reason [reason #f]) (kill pid reason))

(define (extend-transition t . more-actions)
  (match t
    [(transition state actions) (transition state (list actions more-actions))]
    [state (transition state more-actions)]))

(define (prefix-transition t . more-actions)
  (match t
    [(transition state actions) (transition state (list more-actions actions))]
    [state (transition state more-actions)]))

;;---------------------------------------------------------------------------
;; Topics and roles

(define (topic-publisher pattern #:virtual? [virtual? #f])
  (topic 'publisher pattern virtual?))

(define (topic-subscriber pattern #:virtual? [virtual? #f])
  (topic 'subscriber pattern virtual?))

(define (co-roles r)
  (case r
    [(publisher) '(subscriber)]
    [(subscriber) '(publisher)]
    [else #f]))

(define (co-topics t)
  (for/list ([co-role (co-roles (topic-role t))])
    (struct-copy topic t [role co-role])))

(define (topic-union . ts)
  (unless (andmap topic? ts)
    (error 'topic-union "Expects topics as arguments, but given ~v" ts))
  (list->set ts))

(define (refine-topic remote-topic new-pattern)
  (struct-copy topic remote-topic [pattern new-pattern]))

(define (roles-intersect? l r)
  (memq l (co-roles r)))

;; Both left and right must be canonicalized.
(define (topic-intersection left right)
  (and (roles-intersect? (topic-role left) (topic-role right))
       (mgu-canonical (freshen (topic-pattern left)) (freshen (topic-pattern right)))))

;; True iff the flow between remote-topic and local-topic should be
;; visible to the local peer. This is the case when either local-topic
;; is virtual (in which case everything is seen) or otherwise if
;; remote-topic is also not virtual.
(define (flow-visible? local-topic remote-topic)
  (or (topic-virtual? local-topic)
      (not (topic-virtual? remote-topic))))

;;---------------------------------------------------------------------------
;; Core virtualizable virtual machine.

(define (make-vm name boot)
  (vm name
      (hash)
      (hash)
      0
      (list (cons -1 (spawn boot #f 'primordial-process)))))

(define (run-vm state)
  (let loop ((remaining-actions (reverse (vm-pending-actions state)))
	     (state (struct-copy vm state [pending-actions '()]))
	     (outbound-actions '()))
    (match remaining-actions
      ['()
       (let ((state (collect-dead-processes state)))
	 (transition state (reverse (if (vm-idle? state)
					outbound-actions
					(cons (yield run-vm) outbound-actions)))))]
      [(cons (cons pid action) rest)
       (match action
	 [(at-meta-level preaction)
	  (let-values (((transformed-preaction state) (transform-meta-action pid preaction state)))
	    (loop rest state (cons transformed-preaction outbound-actions)))]
	 [(yield k)
	  (loop rest (run-ready state pid k) outbound-actions)]
	 [preaction
	  (let-values (((new-outbound-actions-rev state) (perform-action pid preaction state)))
	    (loop rest state (append new-outbound-actions-rev outbound-actions)))])])))

(define (vm-idle? state)
  (null? (vm-pending-actions state)))

(define (collect-dead-processes state)
  ;; dns-read-driver is being collected because it only has a metarole.%%%
  (define (process-alive? pid p)
    (or (not (set-empty? (process-endpoints p)))
	(not (set-empty? (process-meta-endpoints p)))
	(ormap (lambda (entry) (= (car entry) pid))
	       (vm-pending-actions state))))
  (struct-copy vm state
	       [processes (for/hash ([(pid p) (in-hash (vm-processes state))]
				     #:when (or (process-alive? pid p)
						(begin (printf "~a PID ~v (~a) garbage-collected~n"
							       (vm-name state)
							       pid
							       (process-name p))
						       #f)))
			    (values pid p))]))

(define (send-to-user failure-proc f . args)
  (with-handlers ([exn:fail? failure-proc])
    (apply f args)))

(define (ensure-topic-union t)
  (cond [(topic? t) (set t)]
	[(set? t) t]
	[else
	 (error 'ensure-topic-union
		"Expected either a single topic or a set of topics; got ~v"
		t)]))

(define (perform-action pid preaction state)
  (match preaction
    [(add-role pre-eid topics hs)
     (values '() (do-subscribe pid pre-eid (ensure-topic-union topics) hs state))]
    [(delete-role pre-eid reason)
     (values '() (do-unsubscribe pid pre-eid reason state))]
    [(send-message body role)
     (values '() (route-and-deliver role body state))]
    [(spawn main k debug-name)
     (values '() (do-spawn pid main k debug-name state))]
    [(kill pid-to-kill reason)
     (do-kill (or pid-to-kill pid) reason state)]))

(define (do-subscribe pid pre-eid topics hs state)
  (cond
   [(hash-has-key? (vm-processes state) pid)
    (define new-eid (eid pid pre-eid))
    (define new-endpoint (endpoint new-eid topics hs))
    (let* ((state (notify-route-additions state new-endpoint))
	   (state (generic-update-process state pid (add-process-eid new-eid)))
	   (state (install-endpoint state new-eid new-endpoint)))
      state)]
   [else state]))

(define (generic-update-process state pid updater)
  (struct-copy vm state [processes (hash-update (vm-processes state) pid updater)]))

(define ((add-process-eid new-eid) p)
  (struct-copy process p [endpoints (set-add (process-endpoints p) new-eid)]))

(define ((add-process-meta-eid new-eid) p)
  (struct-copy process p [meta-endpoints (set-add (process-meta-endpoints p) new-eid)]))

(define ((remove-process-eid old-eid) p)
  (struct-copy process p [endpoints (set-remove (process-endpoints p) old-eid)]))

(define ((remove-process-meta-eid old-eid) p)
  (struct-copy process p [meta-endpoints (set-remove (process-meta-endpoints p) old-eid)]))

(define (install-endpoint state new-eid new-endpoint)
  (struct-copy vm state [endpoints (hash-set (vm-endpoints state) new-eid new-endpoint)]))

(define (uninstall-endpoint state old-eid)
  (struct-copy vm state [endpoints (hash-remove (vm-endpoints state) old-eid)]))

(define (notify-route-additions state new-endpoint)
  (match-define (endpoint (eid pid _) topics (handlers presence-handler _ _)) new-endpoint)
  (for*/fold ([state state])
      ([(matching-pid p) (in-hash (vm-processes state))]
       [matching-eid (in-set (process-endpoints p))]
       [e (in-value (hash-ref (vm-endpoints state) matching-eid))]
       [matching-topics (in-value (endpoint-topics e))]
       [matching-topic (in-set matching-topics)]
       [topic (in-set topics)]
       [flow-pattern (in-value (topic-intersection topic matching-topic))]
       #:when flow-pattern)
    (define inbound-flow (refine-topic matching-topic flow-pattern))
    (define outbound-flow (refine-topic topic flow-pattern))
    (define e-presence-handler (handlers-presence (endpoint-handlers e)))
    (let* ((state (if (flow-visible? topic inbound-flow)
		      (run-trapk state pid presence-handler inbound-flow)
		      state))
	   (state (if (flow-visible? matching-topic outbound-flow)
		      (run-trapk state matching-pid e-presence-handler outbound-flow)
		      state)))
      state)))

(define (do-unsubscribe pid pre-eid reason state)
  (define old-eid (eid pid pre-eid))
  (cond
   [(hash-has-key? (vm-endpoints state) old-eid)
    (define old-endpoint (hash-ref (vm-endpoints state) old-eid))
    (let* ((state (generic-update-process state pid (remove-process-eid old-eid)))
	   (state (uninstall-endpoint state old-eid))
	   (state (notify-route-deletions state old-endpoint reason)))
      state)]
   [else state]))

(define (notify-route-deletions state old-endpoint reason)
  (define removed-topics (endpoint-topics old-endpoint))
  (for*/fold ([state state])
      ([(matching-pid p) (in-hash (vm-processes state))]
       [matching-eid (in-set (process-endpoints p))]
       [e (in-value (hash-ref (vm-endpoints state) matching-eid))]
       [matching-topics (in-value (endpoint-topics e))]
       [matching-topic (in-set matching-topics)]
       [removed-topic (in-set removed-topics)]
       [flow-pattern (in-value (topic-intersection removed-topic matching-topic))]
       #:when flow-pattern)
    (define outbound-flow (refine-topic removed-topic flow-pattern))
    (define absence-handler (handlers-absence (endpoint-handlers e)))
    (run-trapk state matching-pid absence-handler outbound-flow reason)))

(define (route-and-deliver role body state)
  (define message-topic (topic role body #f))
  (define endpoints
    (for*/set ([(matching-pid p) (in-hash (vm-processes state))]
	       [matching-eid (in-set (process-endpoints p))]
	       [e (in-value (hash-ref (vm-endpoints state) matching-eid))]
	       [matching-topics (in-value (endpoint-topics e))]
	       [matching-topic (in-set matching-topics)]
	       [flow-pattern (in-value (topic-intersection message-topic matching-topic))]
	       #:when flow-pattern)
      e))
  (for/fold ([state state]) ([e (in-set endpoints)])
    (match-define (endpoint (eid pid _) _ (handlers _ _ message-handler)) e)
    (run-trapk state pid message-handler message-topic body)))

(define (do-spawn spawning-pid main k debug-name state)
  (define new-pid (vm-next-process-id state))
  (define new-name (or debug-name new-pid))
  (match-define (transition initial-state initial-actions)
    (cond
     [(procedure? main) (send-to-user (lambda (e) (transition #f (kill #f e))) main new-pid)]
     [(transition? main) main]))
  (define initial-process (process new-name new-pid initial-state (set) (set)))
  (define spawned-state
    (struct-copy vm (enqueue-actions state new-pid initial-actions)
		 [processes (hash-set (vm-processes state) new-pid initial-process)]
		 [next-process-id (+ new-pid 1)]))
  (printf "~a PID ~v (~a) started~n" (vm-name state) new-pid new-name)
  (flush-output (current-output-port))
  (run-trapk spawned-state spawning-pid k new-pid))

(define (print-kill vm-name pid-to-kill process-name reason)
  (cond
   [(eq? reason #f) (printf "~a PID ~v (~a) exited normally~n"
			    vm-name
			    pid-to-kill
			    process-name)]
   [(exn? reason) ((error-display-handler)
		   (format "~a PID ~v (~a) exited with exception~n~a"
			   vm-name
			   pid-to-kill
			   process-name
			   (exn-message reason))
		   reason)]
   [else (printf "~a PID ~v (~a) exited with reason: ~a~n"
		 vm-name
		 pid-to-kill
		 process-name
		 reason)])
  (flush-output (current-output-port)))

(define (do-kill pid-to-kill reason state)
  (cond
   [(hash-has-key? (vm-processes state) pid-to-kill)
    (define dying-process (hash-ref (vm-processes state) pid-to-kill))
    (print-kill (vm-name state) pid-to-kill (process-name dying-process) reason)
    (let* ((state (for/fold ([state state]) ([eid (in-set (process-endpoints dying-process))])
		    (do-unsubscribe pid-to-kill (eid-pre-eid eid) reason state)))
	   (new-outbound-actions (for/list ([eid (in-set (process-meta-endpoints dying-process))])
				   (delete-role eid reason))))
      (values new-outbound-actions
	      (struct-copy vm state [processes (hash-remove (vm-processes state) pid-to-kill)])))]
   [else (values '() state)]))

(define (run-trapk state pid trap-k . args)
  (if trap-k
      (let ((failure-proc (lambda (e) (lambda (process-state)
					(transition process-state (kill #f e))))))
	(run-ready state pid (apply send-to-user failure-proc trap-k args)))
      state))

(define (maybe-transition->transition t)
  (cond [(transition? t) t]
	[else (transition t '())]))

(define (run-ready state pid interrupt-k)
  (define old-state (process-state (hash-ref (vm-processes state) pid)))
  (match-define (transition new-state actions)
    (maybe-transition->transition
     (send-to-user (lambda (e) (transition old-state (kill #f e))) interrupt-k old-state)))
  (generic-update-process (enqueue-actions state pid actions)
			  pid
			  (lambda (p) (struct-copy process p [state new-state]))))

(define (preaction? a)
  (or (add-role? a)
      (delete-role? a)
      (send-message? a)
      (spawn? a)
      (kill? a)))

(define (action? a)
  (or (preaction? a)
      (yield? a)
      (and (at-meta-level? a)
	   (preaction? (at-meta-level-preaction a)))))

(define (valid-action? pid a)
  (cond
   [(action? a)]
   [(eq? a #f) #f] ;; skip falses in action ConsTrees
   [(void? a) #f] ;; skip voids in action ConsTrees
   [else (log-warning (format "Illegal action ~v from pid ~v" a pid))
	 #f]))

(define (enqueue-actions state pid actions)
  (define flat-actions (for/list ([a (flatten actions)] #:when (valid-action? pid a)) (cons pid a)))
  (struct-copy vm state
	       [pending-actions (append (reverse flat-actions) (vm-pending-actions state))]))

(define (((wrap-trapk pid trapk) . args) state)
  (if (hash-has-key? (vm-processes state) pid)
      (run-vm (apply run-trapk state pid trapk args))
      state))

(define (transform-meta-action pid preaction state)
  (match preaction
    [(add-role pre-eid topics hs)
     (define new-eid (eid pid pre-eid))
     (values (add-role new-eid
		       topics
		       (handlers (wrap-trapk pid (handlers-presence hs))
				 (wrap-trapk pid (handlers-absence hs))
				 (wrap-trapk pid (handlers-message hs))))
	     (if (hash-has-key? (vm-processes state) pid)
		 (generic-update-process state pid (add-process-meta-eid new-eid))
		 state))]
    [(delete-role pre-eid reason)
     (define old-eid (eid pid pre-eid))
     (values (delete-role old-eid reason)
	     (if (hash-has-key? (vm-processes state) pid)
		 (generic-update-process state pid (remove-process-meta-eid old-eid))
		 state))]
    [(? send-message? p)
     (values p state)]
    [(spawn main k debug-name)
     (values (spawn main (wrap-trapk pid k) debug-name) state)]
    [(? kill? p)
     (values p state)]))

(define (nested-vm name boot)
  (lambda (self-pid) (run-vm (make-vm name boot))))

(define (ground-vm boot)
  (let loop ((state (make-vm 'ground-vm boot)))
    (match (let ((t (run-vm state))) (set! current-ground-transition t) t)
      [(transition state actions)
       (define is-blocking?
	 (match actions
	   ['() #t] ;; no "yield" action -> certainly blocking
	   [(list (yield _)) #f] ;; single "yield", with k statically known to be run-vm -> poll
	   [_ (error 'ground-vm
		     "Cannot process meta-actions ~v because no further metalevel exists"
		     actions)]))
       (define active-events
	 (for*/fold ([acc '()])
	     ([(eid e) (in-hash (vm-endpoints state))]
	      [active-topic (in-set (endpoint-topics e))])
	   (match active-topic
	     [(topic 'subscriber (cons (? evt? evt) _) #f)
	      (define ((evt-handler message) state)
		(route-and-deliver 'publisher (cons evt message) state))
	      (cons (wrap-evt evt evt-handler) acc)]
	     [_ acc])))
       (if (and is-blocking? (null? active-events))
	   'done ;; Not polling, and no events that could wake us from blocking, so quit
	   (let ((interruptk (apply sync
				    (if is-blocking?
					never-evt
					(wrap-evt always-evt (lambda (dummy) values)))
				    active-events)))
	     (loop (interruptk state))))])))

(define current-ground-transition #f)